/* partialy inspired by smbmnt.c */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <fcntl.h>
#include <mntent.h>
#include <sys/utsname.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#define SHELL		"/bin/sh"
#define CMD_MAX		1024
#define EXEC_MAX	20

#if DEBUG_LEVEL >= 3
#	define MOUNT_DEBUG
#endif

static void usage(void)
{
	fprintf(stderr, "Usage: shfsmount [opts] [user@]host[:port][/root] mount_point [-o mount_opts]\n\n"
		        "Mount remote file system using shell session.\n\n"
			"opts:\n"
			"  -r\t\tUse rsh (insted of ssh).\n"
			"  -O\t\tSpecify ssh option (e.g. -O \"-c 3des\" == ssh -c 3des)\n"
			"  -u\t\tExecute ssh as this user (root only)\n"
			"  -v\t\tPrint version number.\n"
			"  user\t\tLog in using this user name.\n"
			"  host\t\tSpecify remote host.\n"
			"  port\t\tConnect to this port (ssh only).\n"
			"  root\t\tMake this directory root for mounted filesystem.\n\n"
			"mount_opts (separated by comma):\n"
			"  nocache\tDisable read-ahead and write-back cache.\n"
			"  generic\tDisable use optimized shell command set.\n"
			"  preserve\tPreserve uid/gid (root only).\n"
			"  uid=id\tOwner of all files/dirs on mounted filesystem (root only).\n"
			"  gid=id\tGroup of all files/dirs on mounted filesystem (root only).\n"
			"  rmode=mode\tRoot dir mode.\n"
			"  suid, dev\tsee mount(8) (root only)\n"
			"  ro, rw, nosuid, nodev, exec, noexec\t see mount(8)\n\n"
			"Report bugs to: qiq@atrey.karlin.mff.cuni.cz\n");
	exit(1);
}

static char *fullpath(char *path)
{
        char full[MAXPATHLEN];

	if (strlen(path) > MAXPATHLEN-1)
		return NULL;
        if (!realpath(path, full))
		return NULL;
	return strdup(full);
}

static int mount_ok(char *dir)
{
	struct stat st;

	if (chdir(dir) != 0)
		return 0;
        if (lstat(".", &st) != 0)
		return 0;
        if (!S_ISDIR(st.st_mode)) {
                errno = ENOTDIR;
                return 0;
        }
        if ((getuid() != 0) && 
	    ((getuid() != st.st_uid) || 
	     ((st.st_mode & S_IRWXU) != S_IRWXU))) {
                errno = EPERM;
                return 0;
        }
        return 1;
}

static int update_mtab(char *mnt, char *dev, char *opts)
{
	int fd;
	FILE *mtab;
	struct mntent ment;

        ment.mnt_fsname = dev;
        ment.mnt_dir = mnt;
        ment.mnt_type = "shfs";
        ment.mnt_opts = opts;
        ment.mnt_freq = 0;
        ment.mnt_passno = 0;

        if ((fd = open(MOUNTED"~", O_RDWR|O_CREAT|O_EXCL, 0600)) == -1) {
                fprintf(stderr, "Can't get "MOUNTED"~ lock file");
                return 0;
        }
        close(fd);
        if ((mtab = setmntent(MOUNTED, "a+")) == NULL) {
                fprintf(stderr, "Can't open " MOUNTED);
                return 0;
        }
        if (addmntent(mtab, &ment) == 1) {
                fprintf(stderr, "Can't write mount entry");
                return 0;
        }
        if (fchmod(fileno(mtab), 0644) == -1) {
                fprintf(stderr, "Can't set perms on "MOUNTED);
                return 0;
        }
        endmntent(mtab);
        if (unlink(MOUNTED"~") == -1) {
                fprintf(stderr, "Can't remove "MOUNTED"~");
                return 0;
        }
	return 1;
}


int main(int argc, char **argv)
{
	int use_rsh = 0;
	char *root = NULL;
	char *mnt;
	char *host;
	char *long_host;
	uid_t uid = -1;
	char *execv[EXEC_MAX+10];
	char opt[CMD_MAX];
	char cmd[CMD_MAX];
	char *c;
	int i, j, res;
	int null;
	int shin[2];
	int shout[2];

	if (argc < 2)
		usage();	

	execv[0] = "ssh";
	j = 1;			/* argv index for ssh/rsh */
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-v")) {
			fprintf(stderr, "shfsmount version %s\n", VERSION);
			exit(0);
		} else if (!strcmp(argv[i], "-r")) {
			execv[0] = "rsh";
			use_rsh = 1;
		} else if ((i+1 < argc) && !strcmp(argv[i], "-O")) {
			char *token;

			token = strtok(argv[i+1], " \t\n");
			while (j < EXEC_MAX && token) {
				execv[j++] = token;
				token = strtok(NULL, " \t\n");
			}
			i++;
		} else if ((i+1 < argc) && !strcmp(argv[i], "-u")) {
			struct passwd *passwd;

			if (getuid() != 0) {
				fprintf(stderr, "Only root can do this!\n");
				exit(1);
			}
			passwd = getpwnam(argv[i+1]);
			if (!passwd)
				passwd = getpwuid(atol(argv[i+1]));
			if (!passwd) {
				fprintf(stderr, "Unknown user: %s\n", argv[i+1]);
				exit(1);
			}
			uid = passwd->pw_uid;
			i++;
		} else
			break;
	}

	if (i+1 >= argc)
		usage();
		
	host = argv[i++];
	long_host = strdup(host);
        if ((c = strchr(host, '@'))) {
                *c = '\0';
		execv[j++] = "-l";
		execv[j++] = host;
                host = c + 1;
	}
	if ((c = strchr(host, '/'))) {
		*c = '\0';
		root = c + 1;
	}
	if ((c = strchr(host, ':'))) {
		if (use_rsh) {
			fprintf(stderr, "Cannot specify port when using rsh.\n");
			exit(1);
		}
		*c = '\0';
		execv[j++] = "-p";
		execv[j++] = c + 1;
	}

	mnt = fullpath(argv[i++]);
	if (!mnt) {
		fprintf(stderr, "Invalid path: %s\n", argv[i-1]);
		exit(1);
	}
	if (!mount_ok(mnt)) {
		fprintf(stderr, "%s: %s\n", mnt, strerror(errno));
		exit(1);
	}

	opt[0] = '\0';
	if (root)
		snprintf(opt, CMD_MAX, ",root=/%.256s", root);

	if (i < argc) {
			char *s;
		if (i+1 < argc && !strcmp(argv[i], "-o")) {
			s = argv[i+1];
			while (s) {
				if ((c = strpbrk(s, ",")))
					*(c++) = '\0';
					
				/* check for root options */
				if ((getuid() != 0) && !(strcmp(s, "preserve") 
				   && strncmp(s, "uid=", 4) && strncmp(s, "gid=", 4) 
				   && strcmp(s, "suid") && strcmp(s, "dev"))) {
					fprintf(stderr, "'%s' used by non-root user!\n", s);
					exit(1);
				}

				if (strlen(opt)+1+strlen(s)+1 < CMD_MAX) {
					strcat(opt, ",");
					strcat(opt, s);
				}
				s = c;
			}
		} else {
			fprintf(stderr, "Invalid option: %s\n", argv[i]);
			exit(1);
		}
	}

	/* ensure fd 0-2 are open */
	do {
		if ((null = open("/dev/null", 0)) < 0) {
			fprintf(stderr, "Cannot open /dev/null\n");
			exit(1);
		}
	} while (null < 3);
	close(null);
		
	pipe(shin);
	pipe(shout);

	switch (fork()) {
	case 0:			 /* child */
		close(0);
		dup(shin[0]);
		close(shin[0]);
		close(shin[1]);
		close(1);
		dup(shout[1]);
		close(shout[1]);
		close(shout[0]);
		/* stderr left open */
		
		execv[j++] = host;
		execv[j++] = SHELL;
		execv[j] = NULL;

		if (uid == -1)
			uid = getuid();
		if (setuid(uid)) {
			fprintf(stderr, "%s: Cannot set uid!\n", argv[0]);
			exit(1);
		}

		if (use_rsh)
			execvp("rsh", execv);
		else
			execvp("ssh", execv);

		fprintf(stderr, "%s: Cannot exec remote shell command!\n", argv[0]);
		exit(1);
	case -1:
		fprintf(stderr, "%s: Cannot fork!\n", argv[0]);
		res = 1;
		goto out;
	}
	
	close(shin[0]);
	close(shout[1]);
	
	snprintf(cmd, CMD_MAX, "fdi=%d,fdo=%d%s", shout[0], shin[1], opt);
	if (strlen(cmd) >= CMD_MAX-1) {
		fprintf(stderr, "%s: Agrs too long!\n", argv[0]);
		res = 1;
		goto out;
	}
	
#ifdef MOUNT_DEBUG
	printf("uid: %d, euid: %d\n", getuid(), geteuid());
	printf("%s\n", cmd);
#endif
	if (mount("none", mnt, "shfs", 0, cmd) < 0) {
		if (errno == ENODEV)
			fprintf(stderr, "shfs filesystem not supported by the kernel\n");
		else
			perror(argv[0]);
		res = 1;
		goto out;
	}
	res = update_mtab(mnt, long_host, *opt == ',' ? opt+1 : opt);

out:
	close(shout[0]);
	close(shin[1]);
	free(long_host);
	free(mnt);
	exit(res);
}

